# Matti's Nightmare

# Copyright (C) 2019 Hannu Viitala
#
# The source code in this file is released under the MIT license.
# Go to http://opensource.org/licenses/MIT for the full license details.
#

# 13.6.2019, 3 rooms, fps: 64.5, free ram:9776b, rom size:179kb
# 13.6.2019, 3 rooms, mirroring, fps: 60, free ram:9776b, rom size:179kb
#  3.7.2019, 5 rooms, startup screen, mem.optimized room def, fps: 54-57, free ram:6416b, rom size:182kb

VERSION="4-7-2019, v0.3"

import gc
#gc.collect()
#print ("Begin, free",gc.mem_free())

import upygame as pygame
import urandom as random
import globals as glob
import data
#gc.collect()
#print ("free1",gc.mem_free())
import rooms
#gc.collect()
#print ("free2",gc.mem_free())
import gameobject as GameObjects
import sprite
import gc
import umachine
from micropython import const
import utils

#gc.collect()
#print ("free3",gc.mem_free())


# This is needed to be able to change global variables (!)

# Constants

# vieport move rect
HERO_VP_MOVE_TLX  = const(40)
HERO_VP_MOVE_TLY  = const(40)
HERO_VP_MOVE_BRX  = const(110-40)
HERO_VP_MOVE_BRY  = const(80-40)

HERO_SPEED_X  = const(1)

# Setup global variables
glob.viewPortX = 0
glob.viewPortY = 0
glob.tileAttributeArray = bytearray(16)

# Start state
startRoomX = 0
startRoomY = 0

glob.roomIndex = 5
glob.heroStartPosX=90
glob.heroStartPosY=60

#glob.roomIndex = 9
##heroStartPosX=30
#heroStartPosX=130
#heroStartPosY=70

#glob.roomIndex = 4
#heroStartPosX=100
#heroStartPosY=8

#glob.roomIndex = 6
#heroStartPosX=100
#heroStartPosY=8
#glob.goldCount = 4

# Create tilemap.
# Note: This has to be at the very beginning, because , for sonme reason, if "print()" was used
# before this it will result garbage when drawing the tilemap
glob.tilemap = pygame.tilemap.Tilemap(glob.TILEMAP_W,glob.TILEMAP_H, data.ramMapPixels)

# Determine if we are running in emulator (or simulator)? 
# The pin Ext2 value is 0 in emulator but not in real HW.
glob.isEmulator = False
joyhatY = umachine.Pin ( umachine.Pin.EXT2, umachine.Pin.ANALOG_IN )
joyhatYValue = joyhatY.value()
#print("joyhatY.value()=",test)
if(joyhatYValue == 0):
    glob.isEmulator = True

screen = pygame.display.set_mode() # full screen
pygame.display.init(False)

# ZX Spectrum palette
pygame.display.set_palette_16bit([
    0,      0x001a, 0xd000, 0xd01a, 0x06a0, 0x06ba, 0xd6a0, 0xd6ba,
    0x0000, 0x001f, 0xf800, 0xf81f, 0x07e0, 0x07ff, 0xffe0, 0xffff
]);


# Initialize sound
glob.sound = pygame.mixer.Sound()
glob.currentBgMusic = "rap2h.raw"
if(glob.isEmulator==False):
    glob.sound.play_from_sd(glob.currentBgMusic)

#########################################
def StartupScreen():

    # Create tilemap.
    # Note: This has to be at the very beginning, because , for some reason, if "print()" was used
    # before this it will result garbage when drawing the tilemap
    stTilemap = pygame.tilemap.Tilemap(42,11, data.mapStartupPixels)
    stTilemap.set_tile(0, 8, 8, data.backgroundSurf0)
    stTilemap.set_tile(9, 8, 8, data.backgroundSurf2)

    # The main loop
    stViewPortX = 0
    stViewPortY = 0
    vx = -1
    vy = 0
    doLoop = True
    delay = 2
    tileListIndex = 1
    bgTileListIndex = 0
    tileSfList = [data.backgroundPixels0, data.backgroundSurf2, data.floorSurf1, data.floorSurf3, data.wallSurf2, data.backgroundPixels4, data.backgroundSurf5, data.floorSurf0, data.goldSurf]
    pygame.draw.set_transparent_color( 0 )
    pygame.draw.set_background_color( 0 )
    showInfo = False
    while doLoop:

        glob.currentFrameTime = umachine.time_ms()
        
        # Read the Pokitto standard keys
        eventtype = pygame.event.poll()
        if eventtype != pygame.NOEVENT:
            if eventtype.type == pygame.KEYDOWN:
                if eventtype.key == pygame.BUT_A:
                    doLoop = False
                if eventtype.key == pygame.BUT_B:
                    showInfo = not showInfo
                    #print("showInfo", showInfo)

        # *** UPDATE ***
        if(showInfo):
            
            # Clear
            pygame.draw.rect(pygame.Rect(0,0,110,88), True, 8);

            # Draw text
            y = 1
            #                         xxxxxxxxxxxxxxxxxxxxx     
            pygame.draw.text(  1,  y,"Matti's Nightmare" , 14)
            y += 10
            pygame.draw.text(  1,  y, VERSION , 1)
            y += 7
            pygame.draw.text(  1,  y,"By Hannu Viitala" , 1)
            y += 10
            pygame.draw.text(  1,  y,"Matti is trapped in" , 15)
            y += 7
            pygame.draw.text(  1,  y,"the nightmare. He" , 15)
            y += 7
            pygame.draw.text(  1,  y,"must collect magic" , 15)
            y += 7
            pygame.draw.text(  1,  y,"items to wake up." , 15)
            y += 10
            pygame.draw.text(  1,  y,"Left/right to move." , 7)
            y += 7
            pygame.draw.text(  1,  y,"A button to jump." , 7)
           
        else:
            delay -= 1
            if(delay<=0):
                delay=2
                
                stViewPortX += vx
                if(stViewPortX>0):
                    stViewPortX = 0
                    vx = -vx
                    
                    # Change tiles
                    tileListIndex += 1
                    if(tileListIndex >= len(tileSfList)):
                        tileListIndex = 0
                        bgTileListIndex += 1
                        if(bgTileListIndex >= len(tileSfList)):
                            bgTileListIndex = 0
                        
                    # Only set tiles if they are not the same
                    if(tileListIndex!=bgTileListIndex):
                        stTilemap.set_tile(9, 8, 8, tileSfList[tileListIndex])
                        stTilemap.set_tile(0, 8, 8, tileSfList[bgTileListIndex])
                    
                elif(stViewPortX<-224):
                    stViewPortX = -224
                    vx = -vx
            
                # *** DRAW ***
        
                # Draw tilemap
                stTilemap.draw(stViewPortX, stViewPortY)
                
                # Draw text
                pygame.draw.text(5+1,44-4+1,"  A : Play    B : Info",8)
                pygame.draw.text(5,44-4,"  A : Play    B : Info",15)
            
        pygame.display.flip()


    # Free mamory
    stTilemap.__del__()
    stTilemap = None

#########################################
def main():

    #gc.collect()
    #print ("main() free",gc.mem_free())
                        
    # setup sprite group
    glob.all_enemies = sprite.Group()
    glob.all_enemies_list = []

    # Create hero
    glob.heroGob = GameObjects.HeroGameObject(
        [data.herox0, data.herox1, data.herox2, data.herox1 ],
        [[-1,0],[0,0],[-1,0],[-1,0]])
    glob.heroGob.wx = glob.heroStartPosX
    glob.heroGob.wy = glob.heroStartPosY
    #glob.heroGob.wy = 8 + startRoomY
    glob.heroGob.animDur = 1


    # Create enemies
    dummySurf = data.clockSurf0
    for i in range(glob.ENEMIES_MAX_COUNT):
        enemyGob = GameObjects.EnemyBounceGameObject([dummySurf], [[0,0]])
        glob.all_enemies.add(enemyGob)
        glob.all_enemies_list.append(enemyGob)
    
    # Reset room
    utils.ResetRoom(glob.roomIndex)

   # loop variables
    vx = 0
    vy = 0

    # The main loop
    exitMain = False
    while (not exitMain):

        glob.currentFrameTime = umachine.time_ms()
        
        #!!HV
        skipGameLoop, exitMain = IsGamewonOrLost()
        if( skipGameLoop ):
            continue

        # Read the Pokitto standard keys
        eventtype = pygame.event.poll()
        if eventtype != pygame.NOEVENT:
            if eventtype.type == pygame.KEYDOWN:
                if eventtype.key == pygame.K_RIGHT:
                    vx = HERO_SPEED_X
                if eventtype.key == pygame.K_LEFT:
                    vx = -HERO_SPEED_X
                #if eventtype.key == pygame.K_DOWN:
                #    vy = HERO_SPEED_X
                if eventtype.key == pygame.BUT_A:
                    glob.heroGob.tryJump = True
                    
                #if eventtype.key == pygame.BUT_C:
                #        gc.collect()
                #        print ("free C2:",gc.mem_free())

            if eventtype.type == pygame.KEYUP:
                if eventtype.key == pygame.K_RIGHT:
                    vx = 0
                if eventtype.key == pygame.K_LEFT:
                    vx = 0
                #if eventtype.key == pygame.K_UP:
                #    vy = 0
                #if eventtype.key == pygame.K_DOWN:
                #    vy = 0
                #if eventtype.key == pygame.BUT_A:
                #    shoot = False

        # Store velocity
        glob.heroGob.vx = vx
        #glob.heroGob.vy = vy
        
        # Execute expired events
        removeItemCount = 0
        for ev in glob.timedEventList:
            if(ev[0] < glob.currentFrameTime):
                print("execute timed event", ev)
                removeItemCount += 1
                if(ev[1] == glob.TIMED_EVENT_STOP_MUSIC):
                    print("execute timed event. Reset music")
                    glob.sound.reset()
                    if(isEmulator==False):
                        glob.sound.play_from_sd("dummy123.raw")
            else:
                break
            
        # Remove all expired event
        for i in range(removeItemCount):
            glob.timedEventList.pop(0)
                   

        # *** UPDATE ***
        
        # rare update
        #if(glob.frameNum%20==0):
        #RoomUpdate()

        # Update hero
        #glob.heroGob.update()
        glob.heroGob.update()
        #heroGob3.update()

        # Move viewport
        MoveViewport()

        # Update enemies
        glob.all_enemies.update()

        # *** DRAW ***

        # Draw tilemap
        glob.tilemap.draw(glob.viewPortX, glob.viewPortY)

        # Draw enemies
        glob.all_enemies.draw(screen)

        # Draw hero
        if(glob.heroGob.visible):
            screen.blit(glob.heroGob.image, glob.heroGob.rect.x, glob.heroGob.rect.y, 0, glob.heroGob.hmirror)

        #
        DrawStatusInfo()

        pygame.display.flip()

        glob.frameNum += 1

        #umachine.wait(100)

#########################################
def DrawStatusInfo():

    # Draw gold bitmap and count
    screen.blit( data.goldSurf, 0, 0)
    screen.blit( data.numberBackgroundSurf, 7,0)
    pygame.draw.text(8,1,str(glob.goldCount),15); # white number

    # Draw heart bitmap and count
    screen.blit( data.heartSurf, 110-16, 0)
    screen.blit( data.numberBackgroundSurf, 110-8,0)
    pygame.draw.text(110-7,1,str(glob.heartCount),15); # white number

    # Draw room name
    pygame.draw.rect(pygame.Rect(0,81,110,7), True, 8);
    pygame.draw.text(10,82,glob.roomParameters2[glob.roomIndex][0],15);
    
    # Collect all freed memory. 
    #gc.collect()
    #pygame.draw.text(10,10,str(gc.mem_free()),15);


#########################################
def MoveViewport():

    # Calculate viewport position
    
    # If the hero is outside the viewport, move the viewport
    hero_vpx = glob.heroGob.rect.x
    dx = HERO_VP_MOVE_TLX - hero_vpx
    vpx = 0
    if(dx < 0):
        # not moved too much left
        vpx = 0  # do not move vp
        dx = HERO_VP_MOVE_BRX - hero_vpx
        if(dx > 0):
            # not moved too much right
            vpx = 0  # do not move viewport
        else:
            vpx = dx # move viewport
    else:
        vpx = dx # move viewport
    hero_vpy = glob.heroGob.rect.y
    dy = HERO_VP_MOVE_TLY - hero_vpy
    vpy = 0
    if(dy < 0):
        # not moved too much up
        vpy = 0
        dy = HERO_VP_MOVE_BRY - hero_vpy
        if(dy > 0):
            # not moved too much down
            vpy = 0  # do not move viewport
        else:
            vpy = dy # move viewport
    else:
        vpy = dy # move viewport

    # Move viewport
    newVpX = glob.viewPortX + vpx
    newVpY = glob.viewPortY + vpy

    # Check that viewport is not out bounds
    if(newVpX>0):
        newVpX=0
    elif(newVpX + glob.TILEMAP_W_PX < glob.SCREEN_W):
        newVpX = glob.SCREEN_W - glob.TILEMAP_W_PX
    if(newVpY>0):
        newVpY=0
    elif(newVpY + glob.TILEMAP_H_PX < glob.SCREEN_H-7):
        newVpY = glob.SCREEN_H-7 - glob.TILEMAP_H_PX
    glob.viewPortX = newVpX
    glob.viewPortY = newVpY

    # update hero's position in the viewport
    glob.heroGob.rect.x = glob.heroGob.wx + glob.viewPortX
    glob.heroGob.rect.y = glob.heroGob.wy + glob.viewPortY


#########################################
def IsGamewonOrLost():

    skipGameLoop = False 
    exitMain = False

    if(glob.isGameLost):
        
        #
        # Read B key keys
        eventtype = pygame.event.poll()
        if eventtype != pygame.NOEVENT:
            if eventtype.type == pygame.KEYDOWN:
                 if eventtype.key == pygame.BUT_B:
               
                    # Restart game
                    glob.heroStartPosX=90
                    glob.heroStartPosY=60
                    glob.roomIndex = 5
                    glob.heartCount = 9
                    glob.goldCount = 0
                    glob.isGameLost = False
                    glob.isGameWon = False
                    glob.collectedItems = []
                    glob.timedEventList = []
                   
                    exitMain = True

        # Draw the game over window
        pygame.draw.rect(pygame.Rect(1,44-12,108,24+8), True, 15)
        pygame.draw.rect(pygame.Rect(3,44-10,104,20+8), True, 8)
        pygame.draw.text(5,44-6,"You did not wake up",15)
        pygame.draw.text(28,44-6+8,"GAME OVER!",15)
        pygame.draw.text(5,44-6+16,"Press B to restart",14)
        pygame.display.flip()
        skipGameLoop = True
    
    if(glob.isGameWon):
        
        # Draw random enemy to a random place on screen
        x = random.getrandbits(7) - 4
        y = random.getrandbits(7) - 4
        screen.blit( data.goldSurf, x, y)

        # Draw the game over window
        pygame.draw.rect(pygame.Rect(1,44-12,108,24), True, 15)
        pygame.draw.rect(pygame.Rect(3,44-10,104,20), True, 8)
        pygame.draw.text(5,44-6,"You woke up! The",10)
        pygame.draw.text(5,44-6+8,"nighmare is over!",10)
        pygame.display.flip()
        
        DrawStatusInfo()
        
        skipGameLoop = True

    return skipGameLoop, exitMain

#####################################



# Start mai
#premain()
while(True):
    StartupScreen()
    main()
